/* James Houghton
 * 06 Dec 2018
 * FTP Server: Filesystem interface
 */

#include "server.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

// Transform a requested path into the actual path on the system
char *genpath(char *dest, const struct ftp_conn_t *c, const char *argv) {
    dest[0] = '.';
    strncpy(dest + 1, c->cwd, c->cwd_length + 1);

    if(argv == NULL || strlen(argv) == 0)
        return dest;

    if(argv[0] == '/') {
        strcpy(dest + 1, argv);
    } else {
        if(c->cwd_length != 1) { // cwd does not end with /
            dest[c->cwd_length + 1] = '/';
            strcpy(dest + 2 + c->cwd_length, argv);
        } else {
            strcpy(dest + 1 + c->cwd_length, argv);
        }
    }

    // Remove extra slashes at the end
    size_t e = strlen(dest);
    while(dest[e - 1] == '/') {
        dest[--e] = '\0';
    }

    char *parts[2048];
    memset(parts, 0x0, 2048 * sizeof(char*));

    char *next = dest;
    size_t i = 0;
    while((next = strchr(next + 1, '/')) != NULL) {
        parts[i++] = next + 1;
        *next = '\0';
    }

    // Parse special sequences
    for(size_t j = 0; j < i;) {
        if(strcmp(parts[j], "") == 0) {
            memcpy(&parts[j], &parts[j + 1], (i - j) * sizeof(char *));
            i--;
        } else if(strcmp(parts[j], ".") == 0) {
            memcpy(&parts[j], &parts[j + 1], (i - j) * sizeof(char *));
            i--;
        } else if(strcmp(parts[j], "..") == 0) {
            if(j < 1) {
                errno = ENOENT;
                return NULL;
            }
            memcpy(&parts[j - 1], &parts[j + 1], (i - j) * sizeof(char *));
            i -= 2;
            j--;
        } else {
            j++;
        }
    }

    char path[PATH_LENGTH];
    memset(path, 0x0, PATH_LENGTH);
    path[0] = '/';
    size_t place = 0;

    // Concatenate the parts back into a filename
    char *prev = NULL;
    for(size_t j = 0; parts[j] != NULL; ++j) {
        path[place] = '/';
        if(parts[j] <= prev)
            continue;
        prev = parts[j];
        strcpy(path + place +  1, prev);
        place += strlen(parts[j]) + 1;
    }

    strcpy(dest + 1, path);

    return dest;
}

int send_file(const struct ftp_conn_t *c, const char *args) {
    // Opens a file, and continuously reads data into a buffer
    // writing the data back out onto the network.

    char fn[PATH_LENGTH];
    if(!genpath(fn, c, args)) {
        return errno;
    }

    int fd = open(fn, O_RDONLY);
    if(fd < 0) {
        return errno;
    }

    // Write file to data socket
    char buffer[1024];
    ssize_t num = 0;
    while((num = read(fd, buffer, 1024)) > 0) {
        write(c->data, buffer, num);
    }
    close(fd);

    return 0;
}

int send_ls(const struct ftp_conn_t *c, const char *args) {
    // TYPE is always ASCII here

    // Opens the /bin/ls process, and continuously reads data into a buffer
    // writing the data back out onto the network.
    char buff[1024];
    char fn[PATH_LENGTH];
    char cmd[PATH_LENGTH + 13];

    if(!genpath(fn, c, args)) {
        return errno;
    }

    struct stat file;
    if(stat(fn, &file) < 0)
        return errno;

    if(!S_ISDIR(file.st_mode)) {
        return EACCES;
    }

    // TODO: Fix remote code execution
    snprintf(cmd,  sizeof(cmd), "/bin/ls -l '%s'", fn);
    FILE *fp = popen(cmd, "r");
    if(fp == NULL) {
        perror("failed to execute /bin/ls");
        return errno;
    }

    ssize_t num = 0;
    while((num = fread(buff, sizeof(char), sizeof(buff), fp)) > 0) {
        for(ssize_t i = 0; i < num; ++i) {
            // On-the-fly CR insertion
            if(buff[i] == '\n') {
                write(c->data, "\r\n", 2);
            } else {
                write(c->data, &buff[i], 1);
            }
        }
    }

    int ex = pclose(fp);
    return ex;
}

int recv_file(const struct ftp_conn_t *c, const char *args) {
    // Reads data from the client continuously into a buffer,
    // writing it to disk.

    char fn[PATH_LENGTH];
    if(!genpath(fn, c, args)) {
        return errno;
    }

    int fd = open(fn, O_WRONLY | O_CREAT | O_TRUNC, 0777);
    if(fd < 0) {
        return errno;
    }

    char buffer[1024];
    ssize_t num = 0;
    while((num = read(c->data, buffer, 1024)) > 0) {
        write(fd, buffer, num);
    }
    close(fd);

    return 0;
}

int ftp_mkdir(const struct ftp_conn_t *c, const char *args) {
    // Call mkdir with all available permissions

    char fn[PATH_LENGTH];
    if(!genpath(fn, c, args)) {
        return errno;
    }

    if(mkdir(fn, 0777) == 0) {
        return 0;
    }

    return errno;
}

int ftp_cwd(struct ftp_conn_t *c, const char *args) {

    char fn[PATH_LENGTH];
    if(!genpath(fn, c, args)) {
        return errno;
    }

    // Stat the file to make sure it's readable and a directory
    struct stat file;
    if(stat(fn, &file) < 0)
        return errno;

    if(!S_ISDIR(file.st_mode)) {
        return EACCES;
    }

    strcpy(c->cwd, fn + 1);
    c->cwd_length = strlen(c->cwd);

    return 0;
}
