/* James Houghton
 * 06 Dec 2018
 * FTP Server: Verb implementations
 */

#include "server.h"
#include "functions.h"
#include <arpa/inet.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include "fs.h"
#include <sys/socket.h>

int CMD_QUIT(struct ftp_conn_t *c, const char *args) {
    // Close connection
    ftp_send(c, "221 Goodbye.");
    return FTP_QUIT;
}

int CMD_SYST(struct ftp_conn_t *c, const char *args) {
    // Send system type
    ftp_send(c, "215 UNIX");
    return FTP_SUCCESS;
}

int CMD_USER(struct ftp_conn_t *c, const char *args) {
    // Send 230.
    // We do not care about username or PASS in this case.
    ftp_send(c, "230 Logged in.");
    return FTP_SUCCESS;
}

int CMD_STRU(struct ftp_conn_t *c, const char *args) {
    // We only support FILE structures
    if(args[0] != 'F') {
        ftp_send(c, "504 Non-file structures not supported.");
        return FTP_ERROR;
    } else {
        return FTP_SUCCESS;
    }
}

int CMD_TYPE(struct ftp_conn_t *c, const char *args) {
    // Explicit disallowal of EBCDIC transfer type,
    // but ASCII is allowed because LIST sends ASCII.
    // RETR and STOR require TYPE I.
    switch(args[0]) {
        case 'I':
            c->ft_type = IMAGE;
            ftp_send(c, "200 TYPE set to I.");
            return FTP_SUCCESS;
        case 'A':
            c->ft_type = ASCII;
            ftp_send(c, "200 TYPE set to A.");
            return FTP_SUCCESS;
        case 'E':
        default:
            ftp_send(c, "504 Requested file transport type not supported.");
            return FTP_ERROR;
    }
}

int CMD_PORT(struct ftp_conn_t *c, const char *args) {
    /* Example: PORT 127,0,0,1,25,50
     * IP: 127.0.0.1
     * Port: (25 << 8) + 50
     */

    // Disect IP address from arguments
    char ip[COMMAND_LENGTH];
    char *ip_end   = ip;

    strncpy(ip_end, args, COMMAND_LENGTH - 1);

    for(unsigned i = 0; i < 3; ++i) {
        ip_end = strchr(ip_end, ',');
        if(!ip_end) {
            ftp_send(c, "501 Argument syntax error");
            return FTP_ERROR;
        }
        *ip_end = '.';
    }
    ip_end = strchr(ip_end, ',');
    if(!ip_end) {
        ftp_send(c, "501 Argument syntax error");
        return FTP_ERROR;
    }
    *ip_end = 0;

    // Build port short from arguments
    char *port_upper = ip_end + 1;
    char *port_lower = strchr(port_upper, ',');
    if(!port_lower) {
        ftp_send(c, "501 Argument syntax error");
        return FTP_ERROR;
    }
    *port_lower = 0;
    port_lower++;

    uint16_t port = 0;
    port |= (atoi(port_upper) & 0xFF) << 8;
    port |= (atoi(port_lower) & 0xFF);

    printf("Setting client addr to %s:%d\n", ip, port);

    // Convert IP address string into network order uint32_t
    inet_pton(AF_INET, ip, &c->data_addr.sin_addr);
    // Convert port to network order
    c->data_addr.sin_port = htons(port);
    c->data_addr.sin_family = AF_INET;

    ftp_send(c, "200 Successfully updated client address.");

    return FTP_SUCCESS;
}

int CMD_MODE(struct ftp_conn_t *c, const char *args) {
    /* STREAM transfer mode only */
    switch(args[0]) {
        case 'S':
            c->mode = STREAM;
            ftp_send(c, "200 Transfer mode set to S.");
            return FTP_SUCCESS;
        case 'B':
        case 'C':
        default:
            ftp_send(c, "504 Requested transfer mode not supported.");
            return FTP_ERROR;
    }
}

int CMD_RETR(struct ftp_conn_t *c, const char *args) {
    // If TYPE I has not been issued, return 451
    if(c->ft_type != IMAGE) {
        ftp_send(c, "451 Request action aborted: local error in processing (please issue 'TYPE I').");
        return FTP_ERROR;
    }

    // Attempt to open a data connection corresponding to last PORT call
    if(open_data_conn(c) == FTP_ERROR) {
        return FTP_ERROR;
    }

    // Send the file
    int st = send_file(c, args);
    if(st != 0)
        send_error(c, st);
    else
        ftp_send(c, "226 Transfer complete.");

    close_data_conn(c);
    c->data = -1;

    return FTP_SUCCESS;
}

int CMD_STOR(struct ftp_conn_t *c, const char *args) {
    // If TYPE I has not been issued, return 451
    if(c->ft_type != IMAGE) {
        ftp_send(c, "451 Request action aborted: local error in processing (please issue 'TYPE I').");
        return FTP_ERROR;
    }

    // Attempt to open a data connection corresponding to last PORT call
    if(open_data_conn(c) == FTP_ERROR) {
        return FTP_ERROR;
    }

    // Receive the file
    int st = recv_file(c, args);
    if(st != 0)
        send_error(c, st);
    else
        ftp_send(c, "226 Transfer complete.");

    close_data_conn(c);
    c->data = -1;

    return FTP_SUCCESS;
}

int CMD_LIST(struct ftp_conn_t *c, const char *args) {
    // Although we are not doing ASCII conversion,
    // we allow LIST when TYPE A has been issued.

    if(open_data_conn(c) == FTP_ERROR) {
        return FTP_ERROR;
    }

    // Send file list
    int st = send_ls(c, args);
    if(st != 0)
        send_error(c, st);
    else
        ftp_send(c, "226 Closing connection.");

    close_data_conn(c);
    return FTP_SUCCESS;
}

int CMD_MKD(struct ftp_conn_t *c, const char *args) {
    int exit = ftp_mkdir(c, args);
    if(exit != 0)
        send_error(c, exit);
    else {
        char ret[1024];
        snprintf(ret, 1024, "257 \"%s\" Directory created.", args);
        ftp_send(c, ret);
    }

    return FTP_SUCCESS;
}

int CMD_PWD(struct ftp_conn_t *c, const char *args) {
    char msg[PATH_LENGTH + 6];
    sprintf(msg, "257 \"%s\"", c->cwd);
    ftp_send(c, msg);
    return FTP_SUCCESS;
}

int CMD_CWD(struct ftp_conn_t *c, const char *args) {
    int exit = ftp_cwd(c, args);
    if(exit != 0)
        send_error(c, exit);
    else {
        char ret[PATH_LENGTH + 25];
        snprintf(ret, PATH_LENGTH + 25, "257 \"%s\" Changed directory.", c->cwd);
        ftp_send(c, ret);
    }

    return FTP_SUCCESS;
}

int CMD_NOOP(struct ftp_conn_t *c, const char *args) {
    ftp_send(c, "200 OK.");
    return FTP_SUCCESS;
}

int CMD_NOIMP(struct ftp_conn_t *c, const char *args) {
    // Pseudo-function for non-implemented functions.
    // They are not the same as unknown functions.
    ftp_send(c, "502 Command not implemented.");
    return FTP_ERROR;
}

// Table of defined FTP verbs and function pointers
const struct ftp_cmd_t VERBS[] = {
    { "QUIT", &CMD_QUIT },
    { "SYST", &CMD_SYST },
    { "USER", &CMD_USER },
    { "STRU", &CMD_STRU },
    { "TYPE", &CMD_TYPE },
    { "PORT", &CMD_PORT },
    { "MODE", &CMD_MODE },
    { "RETR", &CMD_RETR },
    { "STOR", &CMD_STOR },
    { "NOOP", &CMD_NOOP },
    { "MKD" , &CMD_MKD  },
    { "LIST", &CMD_LIST },
    { "PWD",  &CMD_PWD  },
    { "CWD",  &CMD_CWD  },

    { "PASS", &CMD_NOIMP },
    { "ACCT", &CMD_NOIMP },
    { "CDUP", &CMD_NOIMP },
    { "SMNT", &CMD_NOIMP },
    { "REIN", &CMD_NOIMP },
    { "PASV", &CMD_NOIMP },
    { "STOU", &CMD_NOIMP },
    { "APPE", &CMD_NOIMP },
    { "ALLO", &CMD_NOIMP },
    { "REST", &CMD_NOIMP },
    { "RNTO", &CMD_NOIMP },
    { "RNFR", &CMD_NOIMP },
    { "ABOR", &CMD_NOIMP },
    { "DELE", &CMD_NOIMP },
    { "RMD",  &CMD_NOIMP },
    { "NLST", &CMD_NOIMP },
    { "SITE", &CMD_NOIMP },
    { "STAT", &CMD_NOIMP },
    { "HELP", &CMD_NOIMP }
};
const size_t NUM_VERBS = \
    sizeof(VERBS) / sizeof(VERBS[0]);
