/* James Houghton
 * 06 Dec 2018
 * FTP Server: Command reader and dispatcher
 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <sys/socket.h>

#include "server.h"
#include "functions.h"

// Map errno.h macros to FTP errors.
int send_error(struct ftp_conn_t *c, int err) {
    // All filesystem functions return errno values,
    // and subsequently call this function to find the proper
    // FTP error code.

    char msg[1024];
    switch(err) {
        case EACCES:
            ftp_send(c, "550 Access denied.");
            break;
        case EBUSY:
            ftp_send(c, "450 Resource busy.");
            break;;
        case ENOENT:
            ftp_send(c, "550 File not found.");
            break;
        case EEXIST:
            ftp_send(c, "521 File exists.");
            break;
        case EROFS:
            ftp_send(c, "550 Read-only file system.");
            break;

        default:
            sprintf(msg, "451 Fatal error: %s", strerror(err));
            ftp_send(c, msg);
            break;

        case 0:
            ftp_send(c, "200 No error. (Server logic error?)");
            break;
    }
    return err;
}

// Find procedure associated with given verb and execute
int ftpexec(struct ftp_conn_t *c, char *command, char* arguments) {
    printf("<----- %s %s\n", command, arguments);
    for(size_t i = 0; i < NUM_VERBS; ++i) {
        if(strcmp(command, VERBS[i].str) == 0) {
            return VERBS[i].func(c, arguments);
        }
    }

    ftp_send(c, "500 Syntax error. Command not recognized");
    return FTP_ERROR;
}

// Send string with CRLF appended
void ftp_send(struct ftp_conn_t *c, const char *msg) {
    printf("-----> %s\n", msg);
    dprintf(c->sock, "%s\r\n", msg);
}

// To be used after open_data_conn()
int close_data_conn(struct ftp_conn_t *c) {
    close(c->data);
    c->data = -1;
    return FTP_SUCCESS;
}

// Opens a data connection with the client
// PORT affects how this data connection is opened.
// close_data_conn() should be run to terminate
// the data connection.
int open_data_conn(struct ftp_conn_t *c) {
    if(c->data < 0) {
        ftp_send(c, "150 File status okay; about to open data connection.");
        c->data = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
        if(c->data < 0) {
            perror("cannot create data socket");
            ftp_send(c, "451 Request action aborted: local error in processing.");
            return FTP_ERROR;
        }
        if(connect(c->data, (struct sockaddr *)&c->data_addr, sizeof c->data_addr) == -1) {
            perror("cannot connect to data socket");
            ftp_send(c, "451 Request action aborted: local error in processing.");
            close(c->data);
            return FTP_ERROR;
        }
    } else {
        ftp_send(c, "125 Data connection already open; transfer starting.");
    }
    return FTP_SUCCESS;
}

// The command parser (separates verb from arguments)
void interp(struct ftp_conn_t *con) {

    ftp_send(con, "220 Welcome!");

    // Keep track of totalread to allow for partial reads
    ssize_t totalread = 0;

    char buffer[COMMAND_LENGTH];
    for(;;) {
        ssize_t numread = read(con->sock, buffer + totalread, sizeof(buffer) - totalread);

        totalread += numread;

        if(numread <= 0) {
            puts("Connection closed unexpectedly");
            break;
        }

        if(buffer[totalread - 1] != '\n')
            continue;

        buffer[totalread - 2] = 0; // remove CRLF

        // Find the command
        char *sp = strchr(buffer, ' ');
        char *args;

        if(sp == NULL) {
            args = NULL; // simply no arguments
        } else {
            *sp = '\0';
            for(char *b = buffer; b < sp; ++b)
                *b = toupper(*b); // uppercase the command
            args = sp + 1; // set arg location
        }

        totalread = 0;

        // Call dispatcher
        switch(ftpexec(con, buffer, args)) {
            case FTP_QUIT:
                goto quit;
            case FTP_ERROR:
            case FTP_SUCCESS:
                continue;
        }
    }
quit:
    return;
}
